Attribute VB_Name = "modParseProc"
'HuntERR Add-In, version 1.0, 07/06/2001
'URFIN JUS (www.urfinjus.net)
'Properties and methods for parsing and analyzing current procedure
Option Explicit

Const STR_STOPONERROR = "If ErrMustStop Then Debug.Assert False: Resume"

Public ProcName As String, procType As vbext_ProcKind, procStartLine As Long, procLineCount As Long
Public procIsHandler As Boolean
Public procText As String, ProcNewText As String
Public procIsSimpleProp As Boolean

Public procLines() As String, procUbnd As Long
Public procHeaderStartLine As Long
Public procTypeStr As String   '"Property", "Function" or "Sub"
Public procHdrStartLine As Long, procHdrEndLine As Long
Public procIsPublic As Boolean
Public procIsExtCallable As Boolean 'True if proc may be called from outside app
Public procParamList As String
Public procErrLabel As String, procErrHStartLine As Long
Public procEndStmtLine  As Long
Public procNumParams As Long

Public procErrHText As String

'Private properties holding information from IDE
Private mProcBrowsable As Boolean, mProcCategory As String, mProcDescr As String
Private mProcHidden As Boolean, mProcDispBind As Boolean, mProcHelpContextID As Long
Private mProcUIDefault As Boolean



Public Sub ScanCurrentProc()
    Dim OperStartLine As Long, OperEndLine As Long, sOper As String, sLabel As String
    On Error GoTo errHandler
    InitProcInfo
    OperStartLine = 0
    While GetOperator(OperStartLine, OperEndLine, sOper, sLabel)
        If TryParseHeader(sOper) Then
            procHdrStartLine = OperStartLine: procHdrEndLine = OperEndLine
        ElseIf ParseToken(STR_ONERRORGOTO, sOper) Then
            procErrLabel = sOper
        Else
            If sLabel <> "" And sLabel = procErrLabel Then procErrHStartLine = OperStartLine
            If StartsWith(sOper, "End " & procTypeStr) Then procEndStmtLine = OperStartLine
        End If
        OperStartLine = OperEndLine + 1
    Wend
    procIsExtCallable = (ideModuleIsExtCallable And procIsPublic) Or procIsHandler
    CheckprocIsSimpleProp
    Exit Sub
errHandler:
    ErrorIn "ScanCurrentProc"
End Sub

Public Sub InitProcInfo()
    procLines = Split(procText, vbNewLine)
    procUbnd = UBound(procLines)
    procErrHStartLine = -1
    procEndStmtLine = -1
    procHdrStartLine = -1
    procHdrEndLine = -1
    procIsSimpleProp = False
End Sub

Private Function GetErrorActionConst() As String
    On Error GoTo errHandler
    If procIsExtCallable And SttUseEAexternCall Then
        GetErrorActionConst = SttConstEAexternCall
    ElseIf (Not procIsExtCallable) And SttUseEAinternCall Then
        GetErrorActionConst = SttConstEAinternCall
    End If
    Exit Function
errHandler:
    ErrorIn "GetErrorActionConst"
End Function

Private Sub CheckprocIsSimpleProp()
    Dim i As Long, TrivStmtCount As Long
    procIsSimpleProp = False
    If procType = vbext_pk_Proc Then Exit Sub
    If procErrHStartLine >= 0 Then Exit Sub 'If there is error handler, it is not simple prop
    If procNumParams > 1 Then Exit Sub
    If Not (procEndStmtLine > procHdrEndLine And procHdrEndLine >= 0) Then Exit Sub
    For i = procHdrEndLine + 1 To procEndStmtLine - 1
        If Not IsCommentOrEmpty(procLines(i)) Then
            If Not IsTrivialAssignment(procLines(i)) Then Exit Sub
            TrivStmtCount = TrivStmtCount + 1
            If TrivStmtCount > 2 Then Exit Sub
        End If
    Next i
    procIsSimpleProp = True
End Sub

Public Sub SetupErrorHandler()
    Dim i As Long
    On Error GoTo errHandler
    'Attach LFs to all lines; we will join them with empty delimiter.
    'This way we can remove lines from code by assigning empty strings
    AttachLFs
    'Delete all "On Error ..." statements
    For i = 0 To procUbnd
        If StartsWith(procLines(i), "On Error ") Then CommentDelete procLines(i), False
    Next i
    'Delete exit operator before error handler if there is one
    If procErrHStartLine > 0 Then
        If StartsWith(procLines(procErrHStartLine - 1), "Exit " & procTypeStr) Then
            CommentDelete procLines(procErrHStartLine - 1), False
        End If
    End If
    'Comment/Delete old error handling block
    If procErrHStartLine >= 0 Then
        For i = procErrHStartLine To procEndStmtLine - 1
            CommentDelete procLines(i), Not SttDeleteOld
        Next i
    End If
    'Set On Error Goto at the beginning
    procLines(procHdrEndLine) = procLines(procHdrEndLine) & Indent & "On Error Goto errHandler" & vbNewLine
    'Set error handling block in place
    If procEndStmtLine >= 0 Then procLines(procEndStmtLine) = procErrHText & procLines(procEndStmtLine)
    ProcNewText = Join(procLines, "")
    Exit Sub
errHandler:
    ErrorIn "SetupErrorHandler"
End Sub

Public Sub ConstructErrorHandlerCode()
    Dim sErrorInCall As String, sEAConst As String, sVarList As String
    Dim sCommas As String 'Temp var to hold commas to add to param list if more params are added
    On Error GoTo errHandler
    sErrorInCall = Indent & "ErrorIn """ & ideModuleName & "." & ProcName
    'Add array of parameters values if necessary
    If procParamList = "" Then
        sErrorInCall = sErrorInCall & """"
        sCommas = " , "
        Else
        If InStr(1, procParamList, ",") > 0 Then
            sVarList = "Array(" & Replace(procParamList, ",", ", ") & ")"
            Else
            sVarList = procParamList
        End If
        sErrorInCall = sErrorInCall & "(" & procParamList & ")"", " & sVarList
        sCommas = ""
    End If
    'Add ErrorAction parameter if necessary
    sEAConst = GetErrorActionConst
    If sEAConst = "" Then
        'Don't add constant, use default value, but remember the comma.
        sCommas = sCommas & ", "
        Else
        sErrorInCall = sErrorInCall & sCommas & ", " & sEAConst
        sCommas = ""
    End If
    sErrorInCall = BreakLine(sErrorInCall, 100, ",")
    procErrHText = Indent & "Exit " & procTypeStr & vbNewLine & _
                "errHandler:" & vbNewLine
    If SttAddStopOnError Then procErrHText = procErrHText & Indent & STR_STOPONERROR & vbNewLine
    procErrHText = procErrHText & sErrorInCall & vbNewLine
    If procIsExtCallable And SttUseEAexternCall And (SttErrResProc <> "") Then
        procErrHText = procErrHText & Indent & SttErrResProc & vbNewLine
    End If
    Exit Sub
errHandler:
    ErrorIn "ConstructErrorHandlerCode"
End Sub



Private Function TryParseHeader(ByVal sHeader As String) As Boolean
    Dim sScope As String
    On Error GoTo errHandler
    If (sHeader = "") Or (procHdrStartLine >= 0) Then Exit Function  'proc header was already found previously
    FindWords "Public ,Private ,Friend", sScope, sHeader
    If Not FindWords("Sub ,Function ,Property ", procTypeStr, sHeader) Then Exit Function
    procTypeStr = Trim$(procTypeStr)
    Select Case sScope
        Case "Public ", "": procIsPublic = True
        Case Else: procIsPublic = False
    End Select
    'Parse parameter list
    ParseParamList sHeader
    TryParseHeader = True
    Exit Function
errHandler:
    ErrorIn "TryParseHeader(sHeader)", sHeader
End Function

Private Sub ParseParamList(ByVal sHeader As String)
    Dim sDecls As String, arrDecls() As String, P1 As Long, P2 As Long, i As Long, sParamName As String
    On Error GoTo errHandler
    procParamList = ""
    P1 = InStr(1, sHeader, "(")
    'Corrected to InStrRev to provide for array parameters
    P2 = InStrRev(sHeader, ")")
    If Not (P2 > P1 And P1 > 0) Then Exit Sub
    sDecls = Trim$(Mid$(sHeader, P1 + 1, P2 - P1 - 1))
    If sDecls = "" Then Exit Sub
    'String default values can contain commas like
    '... Optional Delim As String = "," ....; We parse decls by commas, so get rid of constants
    EmptyStringConstants sDecls
    arrDecls = Split(sDecls, ",")
    For i = LBound(arrDecls) To UBound(arrDecls)
       ParseParamName arrDecls(i), sParamName
       arrDecls(i) = sParamName
    Next i
    procNumParams = UBound(arrDecls) + 1
    procParamList = Join(arrDecls, ",")
    Exit Sub
errHandler:
    ErrorIn "ParseParamList(sHeader)", sHeader
End Sub

'Retrieves multi-line VB operator as one string; returns label in separate string
Private Function GetOperator(ByRef StartLine As Long, ByRef EndLine As Long, _
        ByRef sOper As String, ByRef sLabel As String) As Boolean
    Dim S As String, Done As Boolean
    On Error GoTo errHandler
    sOper = ""
    If StartLine > procUbnd Then Exit Function
    SkipCommentsAndEmpty StartLine
    EndLine = StartLine
    sOper = ""
    Done = False
    Do
        S = Trim$(procLines(EndLine))
        If EndsWith(S, " _") Then
            S = Left$(S, Len(S) - 2)
            EndLine = EndLine + 1
            Else
            Done = True
        End If
        sOper = sOper & S & " "
    Loop Until Done Or (EndLine > procUbnd)
    sOper = Replace(sOper, vbTab, " ")
    ParseLabel sOper, sLabel
    GetOperator = (sOper <> "") Or (sLabel <> "")
    Exit Function
errHandler:
    ErrorIn "GetOperator(StartLine, EndLine, sOper, sLabel)", Array(StartLine, EndLine, sOper, sLabel)
End Function

Private Function SkipCommentsAndEmpty(ByRef LineNum As Long) As Boolean
    While (LineNum < procUbnd) And IsCommentOrEmpty(procLines(LineNum))
        LineNum = LineNum + 1
    Wend
    SkipCommentsAndEmpty = (LineNum <= procUbnd)
End Function

'Attach LF to all lines, so we will join it with empty string as delimiter.
'This way we can remove lines by setting them to empty string.
Private Sub AttachLFs()
    Dim i As Long
    For i = 0 To procUbnd - 1
        procLines(i) = procLines(i) & vbNewLine
    Next i
End Sub

Public Sub SaveCurrProcProps()
    On Error GoTo errHandler
    With ActiveModule.Members(ProcName)
        mProcBrowsable = .Browsable
        mProcCategory = .Category
        mProcDescr = .Description
        mProcHidden = .Hidden
        mProcDispBind = .DisplayBind
        mProcHelpContextID = .HelpContextID
        mProcUIDefault = .UIDefault
    End With
    Exit Sub
errHandler:
    'ErrorIn "modParseProc.SaveCurrProcProps"
End Sub

Public Sub RestoreCurrProcProps()
    On Error GoTo errHandler
    With ActiveModule.Members(ProcName)
        .Browsable = mProcBrowsable
        .Category = mProcCategory
        .Description = mProcDescr
        .Hidden = mProcHidden
        .DisplayBind = mProcDispBind
        .HelpContextID = mProcHelpContextID
        .UIDefault = mProcUIDefault
    End With
    Exit Sub
errHandler:
    'ErrorIn "modParseProc.RestoreCurrProcProps"
End Sub


